import Flutter
import UIKit
import CoreLocation

public class SwiftTcgLocationPlugin: NSObject, FlutterPlugin {
    
   var _callbackChannel: FlutterMethodChannel?
   var _headlessRunner: FlutterEngine?
   var _callbackMapping: String = "location_callback_mapping"
   var _registrar: FlutterPluginRegistrar?
   var _registerPlugins: FlutterPluginRegistrantCallback?
   var initialized: Bool = false
   var backgroundIsolateRun: Bool = false
   var dateFormatter: DateFormatter = DateFormatter()
   let defaultInterval: Int = 30

    public init(registrar: FlutterPluginRegistrar) {
        super.init()
        
        dateFormatter.dateFormat = "yyyy-MM-dd HH:mm:ss"
        
        let channel = FlutterMethodChannel(name: "tcg_location", binaryMessenger: registrar.messenger())
        _headlessRunner = FlutterEngine(name: "LocationIsolate", project: nil, allowHeadlessExecution: true)
        
        if let runner = _headlessRunner as? FlutterBinaryMessenger {
            _callbackChannel = FlutterMethodChannel(name: "tcg_location_background", binaryMessenger: runner)
        }
        
        _registrar = registrar
        
        registrar.addMethodCallDelegate(self, channel: channel)
    }
    
    public static func register(with registrar: FlutterPluginRegistrar) {
       let instance = SwiftTcgLocationPlugin(registrar: registrar)
       registrar.addApplicationDelegate(instance)
    }

    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
            
        print("Method call invoked: \(call.method)")
        
        switch call.method {
        case "initialize":
            // Init
            result(initialize((call.arguments as! [Any?]).first as! CLongLong))
        case "startLocation":
            result(initLocation((call.arguments as! [Any?]).first as! CLongLong))
        case "stopLocation":
            result(stopLocation())
        case "getSampleLocation":
            result("52.403435|4.804436")
        case "getPlatformVersion":
            result("iOS " + UIDevice.current.systemVersion)
        case "initialized":
            result(true)
        case "didLaunchFromBackground":
            result(didLaunchFromBackground())
        default:
            result(FlutterMethodNotImplemented)
        }
    }

    public func application(_ application: UIApplication, didFinishLaunchingWithOptions launchOptions: [AnyHashable : Any] = [:]) -> Bool {
            
        if (launchOptions[UIApplication.LaunchOptionsKey.location] != nil) {
            

            UserDefaults.standard.set(true, forKey: "didLaunchFromBackground")
            UserDefaults.standard.synchronize()
            
            let _ = initialize(getCallbackDispatchHandle())
        }
        
        return true
    }
    
    public static func setPluginRegistrantCallback(_ callback: @escaping FlutterPluginRegistrantCallback) {
        //_registerPlugins = callback
    }

    // MARK: - Custom Location Code
    private lazy var locationManager: CLLocationManager = {
      let manager = CLLocationManager()
      //manager.desiredAccuracy = kCLLocationAccuracyBest
      manager.desiredAccuracy = kCLLocationAccuracyNearestTenMeters
      manager.delegate = self
      manager.requestAlwaysAuthorization()
        
        if #available(iOS 9.0, *) {
            manager.allowsBackgroundLocationUpdates = true
        } else {
            // Fallback on earlier versions
        }
        
      return manager
        
    }()
    
    public func initialize(_ handle: CLongLong) -> Bool {
        
        setCallbackDispatchHandle(handle)
        
        if let info = FlutterCallbackCache.lookupCallbackInformation(handle) {
            let entryPoint = info.callbackName
            let uri = info.callbackLibraryPath
            
            _headlessRunner?.run(withEntrypoint: entryPoint, libraryURI: uri)
        }
        
        if (!backgroundIsolateRun) {
            
            if let runner = _headlessRunner {
                _registerPlugins?(runner)
            }
        }
        
        if let channel = _callbackChannel {
            _registrar?.addMethodCallDelegate(self, channel: channel)
        }
        
        backgroundIsolateRun = true
        
        return true
    }
    
    public func initLocation(_ handle: CLongLong) -> Bool {
        
        setCallbackHandleForLocationUpdate(handle)
        
        locationManager.startUpdatingLocation()
        
        return true
    }

    public func stopLocation() -> Bool {
        
        locationManager.stopUpdatingLocation()
            
        return true
    }
}

// MARK: - CLLocationManagerDelegate
extension SwiftTcgLocationPlugin: CLLocationManagerDelegate {
    
    public func locationManager(_ manager: CLLocationManager, didUpdateLocations locations: [CLLocation]) {
        guard let mostRecentLocation = locations.last else {
          return
        }
        
        let interval = getTimeoutInterval()
        
        if let lastSubmitted = getLastSubmittedDate() {
            let date = dateFormatter.date(from: lastSubmitted)
            
            if let nextDate = Calendar.current.date(byAdding: .second, value: interval, to: date!) {

                if Date() >= nextDate {

                    let callbackHandle = getCallbackHandleForLocationUpdate()
                    _callbackChannel?.invokeMethod("", arguments: [callbackHandle, "\(mostRecentLocation.coordinate.latitude)|\(mostRecentLocation.coordinate.longitude)"])
                    setLastSubmittedDate(dateFormatter.string(from: Date()))
                }
            }
            
        } else {

            let callbackHandle = getCallbackHandleForLocationUpdate()
            _callbackChannel?.invokeMethod("", arguments: [callbackHandle, "\(mostRecentLocation.coordinate.latitude)|\(mostRecentLocation.coordinate.longitude)"])
            setLastSubmittedDate(dateFormatter.string(from: Date()))
        }
    }
    
    public func locationManager(_ manager: CLLocationManager, didFailWithError error: Error) {
        
    }
}

// MARK: - Local State Methods
extension SwiftTcgLocationPlugin {
    
    func setCallbackHandleForLocationUpdate(_ handle: CLongLong) {
        UserDefaults.standard.set(handle, forKey: "location_callback_handle")
        UserDefaults.standard.synchronize()
    }
    
    func getCallbackHandleForLocationUpdate() -> CLongLong {
        if let handle = UserDefaults.standard.object(forKey: "location_callback_handle") as? CLongLong {
            return handle
        }
        
        return 0
    }
    
    func setCallbackDispatchHandle(_ handle: CLongLong) {
        UserDefaults.standard.set(handle, forKey: "callback_dispatcher_handle")
        UserDefaults.standard.synchronize()
    }
    
    func getCallbackDispatchHandle() -> CLongLong {
        if let handle = UserDefaults.standard.object(forKey: "callback_dispatcher_handle") as? CLongLong {
            return handle
        }
        
        return 0
    }
    
    public func setLastSubmittedDate(_ date: String) {

        UserDefaults.standard.set(date, forKey: "last_submitted_date")
        UserDefaults.standard.synchronize()
    }
    
    public func getLastSubmittedDate() -> String? {
        
        if let interval = UserDefaults.standard.object(forKey: "last_submitted_date") as? String {
            return interval
        }
        
        return nil
    }
    
    public func setTimeoutInterval(_ seconds: Int) -> Bool {

        UserDefaults.standard.set(true, forKey: "interval")
        UserDefaults.standard.synchronize()
        
        return true
    }
    
    public func getTimeoutInterval() -> Int {
        
        if let interval = UserDefaults.standard.object(forKey: "interval") as? Int {
            return interval
        }
        
        return defaultInterval
    }
    
    public func didLaunchFromBackground() -> String {
        
        if let didLaunch = UserDefaults.standard.object(forKey: "didLaunchFromBackground") as? Bool {
            
            if didLaunch {
                return "Did launch from background"
            } else {
                return "Did not launch from background"
            }
        }
        
        return "Key not set"
    }
}




